import { MathUtil } from './math.js';
import { Point } from "./point.js";
export var Rect;
(function (Rect) {
    function copy(r) {
        return { pos: Point.copy(r.pos), size: Point.copy(r.size) };
    }
    Rect.copy = copy;
    function area(r) {
        return r.size.x * r.size.y;
    }
    Rect.area = area;
    function isPositive(r) {
        return Point.isPositive(r.size);
    }
    Rect.isPositive = isPositive;
    function closestCorner(r, p) {
        const pow = Math.pow;
        const sqrt = Math.sqrt;
        const near = r.pos;
        const far = { x: r.pos.x + r.size.x, y: r.pos.y + r.size.y };
        const tl = sqrt(pow(p.x - near.x, 2) + pow(p.y - near.y, 2));
        const tr = sqrt(pow(p.x - far.x, 2) + pow(p.y - near.y, 2));
        const br = sqrt(pow(p.x - far.x, 2) + pow(p.y - far.y, 2));
        const bl = sqrt(pow(p.x - near.x, 2) + pow(p.y - far.y, 2));
        const corners = [tl, tr, bl, br];
        // log("Closest corners: " + JSON.stringify(corners));
        corners.sort(function (a, b) { return a - b; });
        // log("Closest corners (sorted): " + JSON.stringify(corners));
        const min = corners[0];
        if (min == tl)
            return 1 /* Anchor.TOPLEFT */;
        if (min == tr)
            return 3 /* Anchor.TOPRIGHT */;
        if (min == br)
            return 5 /* Anchor.BOTTOMRIGHT */;
        return 7 /* Anchor.BOTTOMLEFT */;
    }
    Rect.closestCorner = closestCorner;
    function move(r, p) {
        return {
            pos: Point.add(r.pos, p),
            size: r.size
        };
    }
    Rect.move = move;
    function moveWithin(rect, bounds) {
        const r = copy(rect);
        if (r.size.x > bounds.size.x)
            r.size.x = bounds.size.x;
        if (r.size.y > bounds.size.y)
            r.size.y = bounds.size.y;
        r.pos.x = MathUtil.between(bounds.pos.x, r.pos.x, (bounds.pos.x + bounds.size.x - r.size.x));
        r.pos.y = MathUtil.between(bounds.pos.y, r.pos.y, (bounds.pos.y + bounds.size.y - r.size.y));
        return r;
    }
    Rect.moveWithin = moveWithin;
    function scaleWithin(rect, size) {
        return moveWithin(rect, { pos: Point.ZERO, size: size });
    }
    Rect.scaleWithin = scaleWithin;
    function midpoint(rect) {
        return Point.add(rect.pos, Point.scaleConstant(0.5, rect.size));
    }
    Rect.midpoint = midpoint;
    function moveWithinAffinity(opts) {
        const abs = Math.abs;
        const { original, bounds } = opts;
        const modified = scaleWithin(opts.modified, bounds);
        function close(a, b) {
            return abs(a - b) < opts.margin;
        }
        function axisDiff(axis) {
            const near = close(0, original.pos[axis]);
            const far = close((bounds[axis]), (original.pos[axis] + original.size[axis]));
            // p("near affinity["+axis+"] = "+near);
            // p("far affinity["+axis+"] = "+far);
            if (near && far)
                return 0; // can't do anything for a window attached to both edges
            if (near) {
                // p("maintaining near affinity");
                return -modified.pos[axis];
            }
            else if (far) {
                // p("maintaining far affinity");
                return (bounds[axis]) - (modified.pos[axis] + modified.size[axis]);
            }
            return 0;
        }
        return move(modified, {
            x: axisDiff("x" /* Axis.x */),
            y: axisDiff("y" /* Axis.y */),
        });
    }
    Rect.moveWithinAffinity = moveWithinAffinity;
    function intersection(a, b) {
        let pos = {
            x: Math.max(a.pos.x, b.pos.x),
            y: Math.max(a.pos.y, b.pos.y),
        };
        let bottomRightA = Point.add(a.pos, a.size);
        let bottomRightB = Point.add(b.pos, b.size);
        let bottomRight = {
            x: Math.min(bottomRightA.x, bottomRightB.x),
            y: Math.min(bottomRightA.y, bottomRightB.y),
        };
        let size = Point.subtract(bottomRight, pos);
        if (Point.isPositive(size)) {
            return { pos, size };
        }
        else {
            return null;
        }
    }
    Rect.intersection = intersection;
})(Rect || (Rect = {}));
